@@ -1,10 +1,13 @@
 from __future__ import annotations
 
+from typing import Literal
+
 from math import ceil
 from pathlib import Path
 from copy import deepcopy
 from functools import partial, wraps
 from collections import deque, namedtuple
+from random import random
 
 import numpy as np
 from tqdm import tqdm
@@ -17,6 +20,7 @@ from torch.nn import Module
 from torch.utils.data import TensorDataset, DataLoader
 from torch.distributions import Categorical, Normal
 from torch.utils._pytree import tree_map
+from torch.nested import nested_tensor
 
 from torch.nn.utils.rnn import pad_sequence
 pad_sequence = partial(pad_sequence, batch_first = True)
@@ -33,6 +37,7 @@ b - batch
 n - sequence
 d - dimension
 a - actions
+sr - state rewards
 """
 
 from ema_pytorch import EMA
@@ -62,6 +67,9 @@ from x_transformers_rl.evolution import (
     LatentGenePool
 )
 
+from x_mlps_pytorch.nff import nFeedforwards, norm_weights_
+from x_mlps_pytorch.ff import Feedforwards
+
 from x_mlps_pytorch.mlp import MLP, create_mlp
 
 # memory tuple
@@ -150,6 +158,10 @@ def frac_gradient(t, frac = 1.):
 def log(t, eps = 1e-20):
     return t.clamp(min = eps).log()
 
+def entropy(logits, eps = 1e-20):
+    prob = logits.softmax(dim = -1)
+    return - (log(prob, eps) * prob).sum(dim = -1)
+
 def softclamp(t, value):
     return (t / value).tanh() * value
 
@@ -179,11 +191,21 @@ def temp_batch_dim(fn):
     return inner
 
 def pack_with_inverse(t, pattern):
+    pack_one = is_tensor(t)
+
+    if pack_one:
+        t = [t]
+
     packed, shapes = pack(t, pattern)
 
     def inverse(out, inv_pattern = None):
         inv_pattern = default(inv_pattern, pattern)
-        return unpack(out, shapes, inv_pattern)
+        out = unpack(out, shapes, inv_pattern)
+
+        if pack_one:
+            out = first(out)
+
+        return out
 
     return packed, inverse
 
@@ -205,29 +227,106 @@ def from_numpy(
 
     return t
 
-# activations
+# schedule related
 
-class ReluSquared(Module):
-    def forward(self, x):
-        return x.relu().square() * x.sign()
+def clamp(num, min_val, max_val):
+    return min(max(num, min_val), max_val)
+
+def linear_schedule(min_step, max_step, max_val):
+    slope = max_val / (max_step - min_step)
+
+    def calc_value(step):
+        return clamp(slope * (step - min_step), 0., max_val)
+
+    return calc_value
 
 # action related
 
+def max_neg_value(t):
+    return -torch.finfo(t.dtype).max
+
+def gumbel_noise(t):
+    noise = torch.rand_like(t)
+    return -log(-log(noise))
+
+def to_nested_tensor_and_inverse(tensors):
+    jagged_dims = tuple(t.shape[-1] for t in tensors)
+
+    tensor = cat(tensors, dim = -1)
+
+    tensor, inverse_batch_pack = pack_with_inverse(tensor, '* n')
+
+    batch = tensor.shape[0]
+
+    tensor = rearrange(tensor, 'b n -> (b n)')
+
+    nt = nested_tensor(tensor.split(jagged_dims * batch), layout = torch.jagged, device = tensor.device, requires_grad = True)
+
+    def inverse(t):
+        t = cat(t.unbind())
+        out = rearrange(t, '(b n) -> b n', b = batch)
+        return inverse_batch_pack(out)
+
+    return nt, inverse, inverse_batch_pack
+
+def nested_sum(t, lens: tuple[int, ...]):
+    # needed as backwards not supported for sum with nested tensor
+
+    batch, device = t.shape[0], t.device
+
+    lens = tensor(lens, device = device)
+    indices = lens.cumsum(dim = -1) - 1
+    indices = repeat(indices, 'n -> b n', b = batch)
+
+    cumsum = t.cumsum(dim = -1)
+    sum_at_indices = cumsum.gather(-1, indices)
+    sum_at_indices = F.pad(sum_at_indices, (1, 0), value = 0.)
+
+    return sum_at_indices[..., 1:] - sum_at_indices[..., :-1]
+
+def nested_argmax(t, lens: tuple[int, ...]):
+    batch, device = t.shape[0], t.device
+
+    t, inverse_batch_pack = pack_with_inverse(t, '* nl')
+
+    t = rearrange(t, 'b nl -> nl b')
+    split_tensors = t.split(lens, dim = 0)
+
+    padded_tensors = pad_sequence(split_tensors, batch_first = True, padding_value = max_neg_value(t))
+    padded_tensors = rearrange(padded_tensors, 'n l b -> b n l')
+
+    out = padded_tensors.argmax(dim = -1)
+
+    return inverse_batch_pack(out)
+
 class SafeEmbedding(Module):
     def __init__(
         self,
         num_tokens,
-        dim
+        dim,
+        num_set_actions = 1
     ):
         super().__init__()
-        self.embed = nn.Embedding(num_tokens, dim)
+        self.embed = nn.Embedding(num_set_actions * num_tokens, dim)
+        self.num_set_actions = num_set_actions
+
+        self.register_buffer('offsets', arange(num_set_actions) * num_tokens, persistent = False)
+
+    def forward(
+        self,
+        actions
+    ):
+        if self.num_set_actions == 1:
+            actions = rearrange(actions, '... -> ... 1')
 
-    def forward(self, actions):
         has_actions = actions >= 0.
         actions = torch.where(has_actions, actions, 0)
+
+        actions = actions + self.offsets
         embeds = self.embed(actions)
-        embeds = where('b n, b n d, ', has_actions, embeds, 0.)
-        return embeds
+        embeds = where('... a, ... a d, ', has_actions, embeds, 0.)
+
+        return reduce(embeds, '... a d -> ... d', 'sum')
 
 class Discrete:
     def __init__(
@@ -251,21 +350,59 @@ class Discrete:
     def entropy(self):
         return self.dist.entropy()
 
+class MultiDiscrete:
+    def __init__(
+        self,
+        raw_actions: Tensor | list[Tensor],
+        bin_lens: list[int] | None = None
+    ):
+        if is_tensor(raw_actions):
+            assert exists(bin_lens)
+            raw_actions = raw_actions.split(bin_lens, dim = -1)
+
+        if not exists(bin_lens):
+            bin_lens = [t.shape[-1] for t in raw_actions]
+
+        self.raw_actions = cat(raw_actions, dim = -1)
+        self.bin_lens = bin_lens
+
+        nested_raw_actions, self.inv_nested, self.inv_batch_pack = to_nested_tensor_and_inverse(raw_actions)
+        self.nested_raw_actions = nested_raw_actions
+
+    def sample(self):
+        noise = gumbel_noise(self.raw_actions)
+        noised_logits = self.raw_actions + noise
+        return nested_argmax(noised_logits, self.bin_lens)
+
+    def log_prob(self, values):
+        prob = self.nested_raw_actions.softmax(dim = -1)
+        log_prob = log(prob)
+        log_prob = self.inv_nested(log_prob)
+
+        offsets = tensor(self.bin_lens, device = values.device).cumsum(dim = -1)
+        offsets = F.pad(offsets, (1, -1), value = 0.)
+
+        indices = values + offsets
+        return log_prob.gather(-1, indices)
+
+    def entropy(self):
+        prob = self.nested_raw_actions.softmax(dim = -1)
+        entropies = (-log(prob) * prob).sum(dim = -1)
+        entropies = rearrange(entropies, '(b nl) -> b nl', nl = len(self.bin_lens))
+        return self.inv_batch_pack(entropies, '* nl')
+
 class Continuous:
     def __init__(
         self,
         raw_actions: Tensor,
         squash = False,
-        eps = 1e-5,
-        log_var_clamp_value = 3.
+        eps = 1e-5
     ):
         raw_actions = rearrange(raw_actions, '... (d muvar) -> ... d muvar', muvar = 2)
         self.raw_actions = raw_actions
 
         mean, log_variance = raw_actions.unbind(dim = -1)
 
-        log_variance = softclamp(log_variance, log_var_clamp_value)
-
         variance = log_variance.exp()
 
         self.mean_variance = stack((mean, variance))
@@ -287,7 +424,7 @@ class Continuous:
 
         return sampled.tanh()
 
-    def log_prob(self, value, eps = 1e-5):
+    def log_prob(self, value, eps = 1e-3):
         log_prob = self.dist.log_prob(value)
 
         if not self.squash:
@@ -311,8 +448,10 @@ class WorldModelActorCritic(Module):
         critic_dim_pred,
         critic_min_max_value: tuple[float, float],
         state_dim,
+        state_pred_num_bins = 50,
         continuous_actions = False,
         squash_continuous = False,
+        num_set_actions = 1,
         frac_actor_head_gradient = 5e-2,
         frac_critic_head_gradient = 5e-2,
         entropy_weight = 0.02,
@@ -324,7 +463,11 @@ class WorldModelActorCritic(Module):
         latent_mapper_depth = 2,
         normalize_advantages = True,
         distributed_normalize = True,
-        norm_advantages_stats_momentum = 0.25 # 1. would mean not to use exponential smoothing
+        norm_advantages_stats_momentum = 0.25, # 1. would mean not to use exponential smoothing
+        actor_ff_depth = 1,
+        critic_ff_depth = 2, # certain paper say critic needs to be larger than actor, although in this setting where both draws and slowly shapes the world model, not sure
+        actor_use_norm_ff = False,
+        critic_use_norm_ff = False,
     ):
         super().__init__()
         self.transformer = transformer
@@ -333,7 +476,7 @@ class WorldModelActorCritic(Module):
         self.reward_embed = nn.Parameter(ones(dim) * 1e-2)
 
         if not continuous_actions:
-            self.action_embeds = SafeEmbedding(num_actions, dim)
+            self.action_embeds = SafeEmbedding(num_actions, dim, num_set_actions)
         else:
             self.action_embeds = nn.Sequential(
                 MLP(num_actions, dim * 2, dim, activation = nn.SiLU()),
@@ -344,7 +487,10 @@ class WorldModelActorCritic(Module):
 
         dim = transformer.attn_layers.dim
 
-        self.to_state_embed = nn.Linear(state_dim, dim)
+        self.to_state_embed = nn.Sequential(
+            MLP(state_dim, dim * 2, dim, activation = nn.SiLU()),
+            nn.RMSNorm(dim),
+        )
 
         # world modeling related
 
@@ -356,13 +502,24 @@ class WorldModelActorCritic(Module):
 
         state_dim_and_reward = state_dim + 1
 
-        to_pred_type_klass = Continuous # will setup classic discrete autoregressive w/ vqvae eventually
+        self.state_hl_gauss_loss = HLGaussLoss(
+            min_value = -5., # since state is normalized, this should be sufficient
+            max_value = 5.,
+            num_bins = state_pred_num_bins,
+            clamp_to_range = True
+        )
 
-        self.to_pred = MLP(
-            dim * 2,
-            dim,
-            to_pred_type_klass.dim_out(state_dim_and_reward),
-            activation = nn.SiLU()
+        self.world_model_sos_token = nn.Parameter(torch.randn(dim) * 1e-2)
+        self.null_next_action_token = nn.Parameter(torch.randn(dim) * 1e-2)
+
+        self.to_pred = nn.Sequential(
+            MLP(
+                dim * 2,
+                dim,
+                state_dim_and_reward * state_pred_num_bins,
+                activation = nn.SiLU()
+            ),
+            Rearrange('... (states_rewards bins) -> ... states_rewards bins', bins = state_pred_num_bins)
         )
 
         # evolutionary
@@ -381,16 +538,19 @@ class WorldModelActorCritic(Module):
 
         # actor critic
 
+        actor_net_klass = partial(nFeedforwards, input_preserve_magnitude = True) if actor_use_norm_ff else partial(Feedforwards, final_norm = True)
+        critic_net_klass = partial(nFeedforwards, input_preserve_magnitude = True) if critic_use_norm_ff else partial(Feedforwards, final_norm = True)
+
         actor_critic_input_dim = dim * 2  # gets the embedding from the world model as well as a direct projection from the state
 
         if evolutionary:
             actor_critic_input_dim += dim
 
-        self.critic_head = MLP(
-            actor_critic_input_dim,
-            dim * 2,
-            critic_dim_pred,
-            activation = nn.SiLU()
+        self.critic_head = critic_net_klass(
+            dim,
+            depth = critic_ff_depth,
+            dim_in = actor_critic_input_dim,
+            dim_out = critic_dim_pred,
         )
 
         # https://arxiv.org/abs/2403.03950
@@ -404,13 +564,15 @@ class WorldModelActorCritic(Module):
 
         action_type_klass = Discrete if not continuous_actions else Continuous
 
-        self.action_head = MLP(
-            actor_critic_input_dim,
-            dim * 2,
-            action_type_klass.dim_out(num_actions),
-            activation = nn.SiLU()
+        self.action_head = actor_net_klass(
+            dim,
+            depth = actor_ff_depth,
+            dim_in = actor_critic_input_dim,
+            dim_out = action_type_klass.dim_out(num_actions) * num_set_actions,
         )
 
+        self.num_set_actions = num_set_actions # refactor later, make it work first
+
         if continuous_actions and squash_continuous:
             action_type_klass = partial(action_type_klass, squash = True)
 
@@ -454,14 +616,21 @@ class WorldModelActorCritic(Module):
         pred,
         real
     ):
-        pred_mean, pred_var = pred[..., :-1, :] # todo: fix truncation scenario
-        return F.gaussian_nll_loss(pred_mean, real[:, 1:], pred_var, reduction = 'none')
+        batch = pred.shape[0]
+
+        pred = rearrange(pred[:, :-1], 'b n sr l -> (b sr) n l')
+        real = rearrange(real, 'b n sr -> (b sr) n')
+
+        loss = self.state_hl_gauss_loss(pred, real, reduction = 'none')
+
+        return reduce(loss, '(b sr) n -> b n', 'sum', b = batch)
 
     def compute_done_loss(
         self,
         done_pred,
         dones
     ):
+        dones = F.pad(dones, (1, 0), value = False)
         return F.binary_cross_entropy(done_pred, dones.float(), reduction = 'none')
 
     def compute_actor_loss(
@@ -556,6 +725,8 @@ class WorldModelActorCritic(Module):
         reward_dropout: bool | Tensor | None = None,
         next_actions = None,
         latent_gene = None,
+        world_model_embed_mult = 1.,
+        cache = None,
         **kwargs
     ):
         batch, device = state.shape[0], self.device
@@ -587,11 +758,23 @@ class WorldModelActorCritic(Module):
 
             sum_embeds = sum_embeds + multiply('b ..., b -> b ...', reward_embeds, (~reward_dropout).float())
 
+        # if not inferencing with cache, prepend the sos token
+
+        inference_with_cache = exists(cache)
+
+        prepend_embeds = None
+        if not inference_with_cache:
+            prepend_embeds = repeat(self.world_model_sos_token, 'd -> b 1 d', b = batch)
+
+        # attend
+
         embed, cache = self.transformer(
             state,
             *args,
             **kwargs,
+            cache = cache,
             sum_embeds = sum_embeds,
+            prepend_embeds = prepend_embeds,
             return_embeddings = True,
             return_intermediates = True
         )
@@ -599,19 +782,25 @@ class WorldModelActorCritic(Module):
         # if `next_actions` from agent passed in, use it to predict the next state + truncated / terminated signal
 
         embed_with_actions = None
+
+        if random() < 0.5:
+            next_actions = None # random dropout of next actions
+
         if exists(next_actions):
+
+            next_actions = pad_at_dim(next_actions, (1, 0), value = 0, dim = 1)
+
             next_action_embeds = self.action_embeds(next_actions)
-            embed_with_actions = cat((embed, next_action_embeds), dim = -1)
+        else:
+            seq_len = embed.shape[1]
+            next_action_embeds = repeat(self.null_next_action_token, 'd -> b n d', b = batch, n = seq_len)
+
+        embed_with_actions = cat((embed, next_action_embeds), dim = -1)
 
         # predicting state and dones, based on agent's action
 
-        state_pred = None
-        dones = None
-
-        if exists(embed_with_actions):
-            raw_state_pred = self.to_pred(embed_with_actions)
-            state_pred = Continuous(raw_state_pred).mean_variance
-            dones = self.to_pred_done(embed_with_actions)
+        state_pred = self.to_pred(embed_with_actions)
+        dones = self.to_pred_done(embed_with_actions)
 
         # actor critic input
 
@@ -632,15 +821,26 @@ class WorldModelActorCritic(Module):
             head_input = cat((head_input, latent_embed), dim = -1)
 
         # actor critic heads living on top of transformer - basically approaching online decision transformer except critic learn discounted returns
+        # wm stands for 'world model'
 
-        actor_embed = frac_gradient(embed, self.frac_actor_head_gradient) # what fraction of the gradient to pass back to the world model from the actor / critic head
+        wm_sched_scale = world_model_embed_mult
+        embed_for_actor_critic = embed * wm_sched_scale
 
-        critic_embed = frac_gradient(embed, self.frac_critic_head_gradient)
+        if not inference_with_cache:
+            # remove sos token if needed
+            embed_for_actor_critic = embed_for_actor_critic[:, 1:]
+
+        actor_embed = frac_gradient(embed_for_actor_critic, self.frac_actor_head_gradient * wm_sched_scale) # what fraction of the gradient to pass back to the world model from the actor / critic head
+
+        critic_embed = frac_gradient(embed_for_actor_critic, self.frac_critic_head_gradient * wm_sched_scale)
 
         # actions
 
         raw_actions = self.action_head(cat((head_input, actor_embed), dim = -1))
 
+        if self.num_set_actions > 1:
+            raw_actions = rearrange(raw_actions, '... (action_sets actions) -> ... action_sets actions', action_sets = self.num_set_actions)
+
         # values
 
         values = self.critic_head(cat((head_input, critic_embed), dim = -1))
@@ -766,6 +966,7 @@ class Agent(Module):
         evolutionary = False,
         evolve_every = 1,
         evolve_after_step = 20,
+        num_set_actions = 1,
         latent_gene_pool: dict = dict(
             dim = 128,
             num_genes_per_island = 3,
@@ -790,7 +991,12 @@ class Agent(Module):
         accelerator: Accelerator | None = None,
         actor_loss_weight = 1.,
         critic_loss_weight = 1.,
-        autoregressive_loss_weight = 1.
+        autoregressive_loss_weight = 1.,
+        world_model_embed_linear_schedule: (
+            float |                 # 1. fully at this step
+            tuple[float, float] |   # min step to this max step from 0. to 1.
+            None
+        ) = None
     ):
         super().__init__()
 
@@ -807,6 +1013,7 @@ class Agent(Module):
 
         self.model = WorldModelActorCritic(
             num_actions = num_actions,
+            num_set_actions = num_set_actions,
             continuous_actions = continuous_actions,
             squash_continuous = squash_continuous,
             critic_dim_pred = critic_pred_num_bins,
@@ -877,6 +1084,18 @@ class Agent(Module):
         self.register_buffer('step', tensor(0))
         self.to(accelerator.device)
 
+        # slowly titrate in the world model embed into actor / critic, as initially it won't provide much value
+
+        self.world_model_sched = None
+
+        if exists(world_model_embed_linear_schedule):
+            embed_mult_range = world_model_embed_linear_schedule
+
+            if isinstance(embed_mult_range, float):
+                embed_mult_range = (0., embed_mult_range)
+
+            self.world_model_sched = linear_schedule(*embed_mult_range, 1.)
+
         # loss weights
 
         self.actor_loss_weight = actor_loss_weight
@@ -887,6 +1106,13 @@ class Agent(Module):
     def device(self):
         return self.step.device
 
+    @property
+    def world_model_embed_mult(self):
+        if not exists(self.world_model_sched):
+            return 1.
+
+        return self.world_model_sched(self.step.item())
+
     def save(self):
         if not self.accelerator.is_main_process:
             return
@@ -1033,7 +1259,8 @@ class Agent(Module):
                     actions = prev_actions,
                     latent_gene = latent_gene,
                     next_actions = actions, # prediction of the next state needs to be conditioned on the agent's chosen action on that state, and will make the world model interactable
-                    mask = mask
+                    mask = mask,
+                    world_model_embed_mult = self.world_model_embed_mult
                 )
 
                 # autoregressive loss for transformer world modeling - there's nothing better atm, even if deficient
@@ -1043,12 +1270,14 @@ class Agent(Module):
                     states_with_rewards
                 )
 
-                world_model_loss = world_model_loss[mask[:, :-1]]
+                world_model_loss = world_model_loss[mask]
 
                 # predicting termination head
 
+                done_pred_mask = F.pad(mask, (1, 0), value = True)
+
                 pred_done_loss = model.compute_done_loss(done_pred, dones)
-                pred_done_loss = pred_done_loss[mask]
+                pred_done_loss = pred_done_loss[done_pred_mask]
 
                 # update actor and critic
 
@@ -1093,6 +1322,10 @@ class Agent(Module):
                 optimizer.step()
                 optimizer.zero_grad()
 
+                # norm any weights
+
+                norm_weights_(self)
+
                 # log losses
 
                 logs = dict(
@@ -1162,7 +1395,7 @@ class Agent(Module):
             cache = hiddens,
         )
 
-        raw_actions = rearrange(raw_actions, '1 1 d -> d')
+        raw_actions = rearrange(raw_actions, '1 1 ... -> ...')
 
         return raw_actions, next_hiddens
 
@@ -1175,13 +1408,17 @@ class Learner(Module):
         num_actions,
         reward_range,
         world_model: dict,
+        num_set_actions = 1,
         continuous_actions = False,
+        discretize_continuous = True,
+        continuous_discretized_bins = 50,
         squash_continuous = True,
         continuous_actions_clamp: tuple[float, float] | None = None,
         evolutionary = False,
         evolve_every = 10,
         evolve_after_step = 20,
         reward_dropout_prob = 0.5,
+        curiosity_reward_weight = 0.,
         latent_gene_pool: dict | None = None,
         max_timesteps = 500,
         batch_size = 8,
@@ -1195,7 +1432,7 @@ class Learner(Module):
         beta_s = .01,
         regen_reg_rate = 1e-4,
         cautious_factor = 0.1,
-        epochs = 4,
+        epochs = 3,
         ema_decay = 0.9,
         save_every = 100,
         accelerate_kwargs: dict = dict(),
@@ -1207,8 +1444,26 @@ class Learner(Module):
 
         self.accelerator = Accelerator(**accelerate_kwargs)
 
+        # if doing continuous actions but want it discretized, have the learning orchestrator take care of converting the discrete to continuous from agent -> env
+
+        # todo - move logic onto Agent so during deployment mapping is taken care of there
+
+        self.discrete_to_continuous = None
+
+        if continuous_actions and discretize_continuous:
+            continuous_actions = False
+            num_set_actions = num_actions  # fix this later, num actions for continuous should be 1
+            num_actions = continuous_discretized_bins
+            assert exists(continuous_actions_clamp)
+
+            min_val, max_val = continuous_actions_clamp
+            self.discrete_to_continuous = torch.linspace(min_val, max_val, continuous_discretized_bins)
+
+        self.num_set_actions = num_set_actions
+
         self.agent = Agent(
             state_dim = state_dim,
+            num_set_actions = num_set_actions,
             num_actions = num_actions,
             continuous_actions = continuous_actions,
             squash_continuous = squash_continuous,
@@ -1274,6 +1529,10 @@ class Learner(Module):
 
         self.reward_dropout_prob = reward_dropout_prob
 
+        # weight for curiosity reward, based on entropy of state prediction
+
+        self.curiosity_reward_weight = curiosity_reward_weight
+
         # saving agent
 
         self.save_every = save_every
@@ -1293,6 +1552,7 @@ class Learner(Module):
         seed = None,
         max_timesteps = None
     ):
+        has_curiosity_reward = self.curiosity_reward_weight > 0.
         max_timesteps = default(max_timesteps, self.max_timesteps)
 
         agent, device, num_episodes_per_update, is_main = self.agent, self.device, self.num_episodes_per_update, self.accelerator.is_main_process
@@ -1359,7 +1619,8 @@ class Learner(Module):
                 if self.continuous_actions:
                     prev_action = zeros((self.num_actions,), device = device)
                 else:
-                    prev_action = tensor(-1).to(device)
+                    prev_action = -1 if self.num_set_actions == 1 else ([-1] * self.num_set_actions)
+                    prev_action = tensor(prev_action, device = device)
 
                 prev_reward = tensor(0.).to(device)
 
@@ -1384,36 +1645,46 @@ class Learner(Module):
                     normed_state = rearrange(normed_state, 'd -> 1 1 d')
                     prev_action = rearrange(prev_action, '... -> 1 1 ...')
 
-                    raw_actions, values, _, _, world_model_cache = model.forward_eval(
+                    raw_actions, values, state_pred, _, world_model_cache = model.forward_eval(
                         normed_state,
                         rewards = normed_reward,
                         reward_dropout = reward_dropout,
                         latent_gene = latent_gene,
                         cache = world_model_cache,
-                        actions = prev_action
+                        actions = prev_action,
+                        world_model_embed_mult = agent.world_model_embed_mult
                     )
 
-                    raw_actions = rearrange(raw_actions, '1 1 d -> d')
+                    raw_actions = rearrange(raw_actions, '1 1 ... -> ...')
                     values = rearrange(values, '1 1 d -> d')
 
-                    return model.action_type_klass(raw_actions), values
+                    state_pred = state_pred[:, -1] # last token, in the case of sos token added
+                    state_pred_entropy = entropy(state_pred).mean(dim = -1)
+                    state_pred_entropy = rearrange(state_pred_entropy, '1 ->')
+
+                    return model.action_type_klass(raw_actions), values, state_pred_entropy
 
                 cumulative_rewards = 0.
 
                 for timestep in range(max_timesteps):
 
-                    dist, value = state_to_pred_action_and_value(state, prev_action, prev_reward, latent_gene)
+                    dist, value, state_pred_entropy = state_to_pred_action_and_value(state, prev_action, prev_reward, latent_gene)
 
                     action = dist.sample()
                     action_log_prob = dist.log_prob(action)
 
-                    if self.continuous_actions and exists(self.continuous_actions_clamp):
+                    action_to_env = action.cpu()
+
+                    if exists(self.discrete_to_continuous):
+                        action_to_env = self.discrete_to_continuous[action_to_env]
+
+                    elif self.continuous_actions and exists(self.continuous_actions_clamp):
                         # environment clamping for now, before incorporating squashed gaussian etc
 
                         clamp_min, clamp_max = self.continuous_actions_clamp
                         action.clamp_(clamp_min, clamp_max)
 
-                    env_step_out = env.step(action.tolist())
+                    env_step_out = env.step(action_to_env.tolist())
 
                     if len(env_step_out) >= 4:
                         next_state, reward, terminated, truncated, *_ = env_step_out
@@ -1426,6 +1697,10 @@ class Learner(Module):
                     next_state = from_numpy(next_state).to(device)
 
                     reward = float(reward)
+
+                    if has_curiosity_reward:
+                        reward += state_pred_entropy.item() * self.curiosity_reward_weight
+
                     cumulative_rewards += reward
 
                     prev_action = action
